# This file is part of Wike (com.github.hugolabe.Wike)
# SPDX-FileCopyrightText: 2021-25 Hugo Olabera
# SPDX-License-Identifier: GPL-3.0-or-later


from gi.repository import Gtk, Adw

from wike.data import languages


# Languages dialog to choose user languages

@Gtk.Template(resource_path='/com/github/hugolabe/Wike/gtk/languages.ui')
class LanguagesDialog(Adw.Dialog):

  __gtype_name__ = 'LanguagesDialog'

  search_bar = Gtk.Template.Child()
  languages_entry = Gtk.Template.Child()
  languages_list = Gtk.Template.Child()
  select_all_button = Gtk.Template.Child()
  select_none_button = Gtk.Template.Child()
  selected_label = Gtk.Template.Child()

  # Initialize and connect signals

  def __init__(self, window):
    super().__init__()

    self._window = window

    self._languages_changed = False
    self._languages_selected = 0
    
    self.search_bar.set_key_capture_widget(self)
    self.languages_list.set_filter_func(self._filter_list)
    self._populate()

    self.languages_entry.connect('search-changed', self._languages_entry_changed_cb)
    self.languages_list.connect('row-activated', self._languages_list_selected_cb)
    self.select_all_button.connect('clicked', self._select_all_button_cb)
    self.select_none_button.connect('clicked', self._select_none_button_cb)
    self.connect('closed', self._dialog_closed_cb)

  # Filter languages list for languages entry content

  def _filter_list(self, row):
    text = self.languages_entry.get_text()
    if text == '':
      return True

    if row.lang_name.lower().startswith(text.lower()) or row.lang_id.lower().startswith(text.lower()):
      return True
    else:
      return False

  # Populate list of available languages

  def _populate(self):
    for lang_id in sorted(languages.wikilangs):
      lang_name = languages.wikilangs[lang_id].capitalize()
      if lang_id in languages.items:
        row = LanguagesRow(lang_name, lang_id, True)
        self._languages_selected += 1
      else:
        row = LanguagesRow(lang_name, lang_id, False)
      self.languages_list.append(row)
      row.lang_check.connect('toggled', self._language_checkbutton_cb)

    self._set_selected_label()

  # Refresh languages list on languages entry changed

  def _languages_entry_changed_cb(self, languages_entry):
    self.languages_list.invalidate_filter()

  # On row selected set language check button

  def _languages_list_selected_cb(self, languages_list, row):
    row.lang_check.set_active(not row.lang_check.get_active())

  # On check button changed update variables

  def _language_checkbutton_cb(self, check_button):
    if check_button.get_active():
      self._languages_selected += 1
    else:
      self._languages_selected -= 1
    self._set_selected_label()

    if not self._languages_changed:
      self._languages_changed = True

  # On button click check all languages

  def _select_all_button_cb(self, select_all_button):
    i = 0
    while True:
      row = self.languages_list.get_row_at_index(i)
      if row:
        row.lang_check.set_active(True)
        i += 1
      else:
        break

  # On button click uncheck all languages

  def _select_none_button_cb(self, select_none_button):
    i = 0
    while True:
      row = self.languages_list.get_row_at_index(i)
      if row:
        row.lang_check.set_active(False)
        i += 1
      else:
        break

  # Update selected languages label
  
  def _set_selected_label(self):
    if self._languages_selected == 1:
      self.selected_label.set_label(_('1 language selected'))
    else:
      self.selected_label.set_label(_('%s languages selected') % self._languages_selected)

  # On dialog closed refresh languages list (if changed)

  def _dialog_closed_cb(self, languages_dialog):
    if self._languages_changed:
      languages.clear()

      i = 0
      while True:
        row = self.languages_list.get_row_at_index(i)
        if row:
          if row.lang_check.get_active():
            languages.items[row.lang_id] = row.lang_name
          i += 1
        else:
          break

      if len(languages.items) == 0:
        languages.items['en'] = 'English'

      self._window.search_panel.settings_popover.populate_list()
      self._window.langlinks_panel.populate(self._window.page.wikiview.langlinks)

    return False


# Row in languages list

@Gtk.Template(resource_path='/com/github/hugolabe/Wike/gtk/languages-row.ui')
class LanguagesRow(Gtk.ListBoxRow):

  __gtype_name__ = 'LanguagesRow'

  name_label = Gtk.Template.Child()
  id_label = Gtk.Template.Child()
  lang_check = Gtk.Template.Child()

  # Set row values

  def __init__(self, lang_name, lang_id, checked):
    super().__init__()

    self.lang_name = lang_name
    self.lang_id = lang_id

    self.name_label.set_label(lang_name)
    self.id_label.set_label(lang_id)
    self.lang_check.set_active(checked)



# This file is part of Wike (com.github.hugolabe.Wike)
# SPDX-FileCopyrightText: 2021-25 Hugo Olabera
# SPDX-License-Identifier: GPL-3.0-or-later


from gi.repository import Gdk, Gtk, Adw, WebKit

from wike.data import settings
from wike.view import WikiView


# Page box for each tab

@Gtk.Template(resource_path='/com/github/hugolabe/Wike/gtk/page.ui')
class PageBox(Gtk.Box):

  __gtype_name__ = 'PageBox'

  search_bar = Gtk.Template.Child()
  search_entry = Gtk.Template.Child()
  search_prev_button = Gtk.Template.Child()
  search_next_button = Gtk.Template.Child()
  search_matches_label = Gtk.Template.Child()
  view_stack = Gtk.Template.Child()

  # Add stack pages, initialize find controller and connect signals

  def __init__(self, window, lazy_load):
    super().__init__()

    self._window = window
    self.lazy_load = lazy_load

    self.wikiview = WikiView()
    self.wikiview.set_vexpand(True)
    self.view_stack.add_named(self.wikiview, 'wikiview')
    self.view_stack.set_visible_child_name('wikiview')

    self.status = PageStatus(self)
    self.view_stack.add_named(self.status, 'status')

    find_controller = self.wikiview.get_find_controller()
    nav_list = self.wikiview.get_back_forward_list()
    self.search_bar.connect_entry(self.search_entry)

    event_controller_key = Gtk.EventControllerKey.new()
    self.wikiview.add_controller(event_controller_key)
    webkit_allow_keys = (0xff09, 0xff0d, 0xff50, 0xff51, 0xff52, 0xff53, 0xff54, 0xff55, 0xff56, 0xff57)
    event_controller_key.connect('key-pressed', self._event_controller_key_pressed_cb, webkit_allow_keys)
    
    self.wikiview.connect('load-changed', self._wikiview_load_changed_cb)
    self.wikiview.connect('load-props', self._wikiview_load_props_cb)
    self.wikiview.connect('new-page', self._wikiview_new_page_cb)
    self.wikiview.connect('add-bookmark', self._wikiview_add_bookmark_cb)
    self.search_entry.connect('search-changed', self._search_entry_changed_cb, find_controller)
    self.search_entry.connect('activate', self._search_entry_activate_cb, find_controller)
    self.search_prev_button.connect('clicked', self._search_prev_button_cb, find_controller)
    self.search_next_button.connect('clicked', self._search_next_button_cb, find_controller)
    find_controller.connect('found-text', self._find_controller_found_cb)
    find_controller.connect('failed-to-find-text', self._find_controller_not_found_cb)
    find_controller.connect('counted-matches', self._find_controller_matches_cb)
    nav_list.connect('changed', self._nav_list_changed_cb)

  # Set focus to visible stack page

  def set_focus(self):
    if self.view_stack.get_visible_child_name() == 'status':
      self.status.main_page_button.grab_focus()
    else:
      self.wikiview.grab_focus()

  # Load a non-loaded page, for example when the user selects the tab

  def load_now(self):
    uri, title = self.lazy_load
    self.wikiview.load_wiki(uri)
    self.lazy_load = False

  # Manage webkit key pressed and redirect to window

  def _event_controller_key_pressed_cb(self, event_controller, keyval, keycode, state, allow_keys):
    if keyval in allow_keys or state is Gdk.ModifierType.CONTROL_MASK:
      return False
    else:
      event_controller.forward(self._window)
      return True

  # Manage wikiview load page events

  def _wikiview_load_changed_cb(self, wikiview, event):
    tabpage = self._window.tabview.get_page(self)

    match event:

      case WebKit.LoadEvent.STARTED:
        self._is_main = wikiview.get_uri().endswith('.wikipedia.org/')
        self.search_bar.set_search_mode(False)
        self.view_stack.set_visible_child_name('wikiview')
        tabpage.set_title(_('Loading…'))
        tabpage.set_loading(True)
        if tabpage.get_selected():
          self._window.set_title(_('Loading…'))
          wikiview.grab_focus()

      case WebKit.LoadEvent.COMMITTED:
        theme = settings.get_int('theme')
        match theme:
          case 1:
            wikiview.set_wikipedia_skin(True)
          case 3:
            style_manager = Adw.StyleManager.get_default()
            if style_manager.get_dark():
              wikiview.set_wikipedia_skin(True)

      case WebKit.LoadEvent.FINISHED:
        tabpage.set_loading(False)
        if wikiview.is_local():
          self.status.show(wikiview.get_uri())

  # On wikiview load page properties populate toc and langlinks
  
  def _wikiview_load_props_cb(self, wikiview):
    tabpage = self._window.tabview.get_page(self)

    tabpage.set_title(wikiview.title)
    if tabpage.get_selected():
      self._window.set_title(wikiview.title)
      self._window.toc_panel.populate(wikiview.title, wikiview.sections)
      self._window.langlinks_panel.populate(wikiview.langlinks)
      self._window.bookmarks_panel.refresh_buttons()
      self._window.refresh_menu_actions(wikiview.is_local())

    if settings.get_boolean('keep-history'):
      if not self._is_main and not wikiview.is_local():
        self._window.history_panel.add_item(wikiview.get_base_uri(), wikiview.title, wikiview.get_lang())

  # On webview event create new page

  def _wikiview_new_page_cb(self, wikiview, uri):
    tabpage = self._window.tabview.get_page(self)
    self._window.new_page(uri, tabpage, False)

  # On webview event add new bookmark

  def _wikiview_add_bookmark_cb(self, wikiview, uri, title, lang):
    if self._window.bookmarks_panel.add_bookmark(uri, title, lang):
      message = _('Bookmark added: ') + title
      self._window.send_notification(message)

  # Search text in article when entry changes

  def _search_entry_changed_cb(self, search_entry, find_controller):
    text = search_entry.get_text()
    if len(text) > 2:
      find_controller.count_matches(text, WebKit.FindOptions.WRAP_AROUND | WebKit.FindOptions.CASE_INSENSITIVE, 9999)
      find_controller.search(text, WebKit.FindOptions.WRAP_AROUND | WebKit.FindOptions.CASE_INSENSITIVE, 9999)
    else:
      self.search_matches_label.set_label('')
      self.search_entry.remove_css_class('error')
      self.search_prev_button.set_sensitive(False)
      self.search_next_button.set_sensitive(False)
      find_controller.search_finish()

  # On entry activated search next match

  def _search_entry_activate_cb(self, search_entry, find_controller):
    text = search_entry.get_text()
    if len(text) > 2:
      find_controller.search_next()

  # On text search prev button clicked search previous match

  def _search_prev_button_cb(self, search_prev_button, find_controller):
    find_controller.search_previous()

  # On text search next button clicked search next match

  def _search_next_button_cb(self, search_next_button, find_controller):
    find_controller.search_next()

  # Found text in article

  def _find_controller_found_cb(self, find_controller, match_count):
    self.search_prev_button.set_sensitive(True)
    self.search_next_button.set_sensitive(True)

  # Not found text in article

  def _find_controller_not_found_cb(self, find_controller):
    self.search_matches_label.set_label('')
    self.search_entry.add_css_class('error')
    self.search_prev_button.set_sensitive(False)
    self.search_next_button.set_sensitive(False)
    find_controller.search_finish()

  # Show text search matches found

  def _find_controller_matches_cb(self, find_controller, match_count):
    if match_count > 0:
      self.search_matches_label.set_label(str(match_count))
      self.search_entry.remove_css_class('error')

  # Refresh navbox state on navlist changed

  def _nav_list_changed_cb(self, nav_list, item_added, item_removed):
    tabpage = self._window.tabview.get_page(self)
    if tabpage.get_selected():
      self._window.refresh_nav_actions(self.wikiview)


# Status page for view stack

@Gtk.Template(resource_path='/com/github/hugolabe/Wike/gtk/page-status.ui')
class PageStatus(Adw.Bin):

  __gtype_name__ = 'PageStatus'

  status_page = Gtk.Template.Child()
  main_page_button = Gtk.Template.Child()
  random_article_button = Gtk.Template.Child()
  try_again_button = Gtk.Template.Child()

  # Initialize and connect signals

  def __init__(self, page):
    super().__init__()

    self._page = page

    self.try_again_button.connect('clicked', self._try_again_button_cb)

  # Show status for provided uri

  def show(self, uri):
    match uri:
      case 'about:notfound':
        self.status_page.set_title(_('Article not Found'))
        self.status_page.set_description(_('Can’t find any results for requested query'))
        self.status_page.set_icon_name('find-location-symbolic')
        self.main_page_button.set_visible(True)
        self.random_article_button.set_visible(False)
        self.try_again_button.set_visible(False)
        self.main_page_button.grab_focus()
      case 'about:error':
        self.status_page.set_title(_('Can’t Access Wikipedia'))
        self.status_page.set_description(_('Check your Internet connection and try again'))
        self.status_page.set_icon_name('network-error-symbolic')
        self.main_page_button.set_visible(False)
        self.random_article_button.set_visible(False)
        self.try_again_button.set_visible(True)
        self.try_again_button.grab_focus()
      case _:
        self.status_page.set_title(_('Search Wikipedia'))
        self.status_page.set_description(_('Start typing to search Wikipedia articles'))
        self.status_page.set_icon_name('com.github.hugolabe.Wike')
        self.main_page_button.set_visible(True)
        self.random_article_button.set_visible(True)
        self.try_again_button.set_visible(False)
        self.main_page_button.grab_focus()

    self._page.view_stack.set_visible_child_name('status')

  # Try again button clicked

  def _try_again_button_cb(self, try_again_button):
    if self._page.wikiview.fail_uri:
      self._page.wikiview.load_wiki(self._page.wikiview.fail_uri)



# This file is part of Wike (com.github.hugolabe.Wike)
# SPDX-FileCopyrightText: 2021-25 Hugo Olabera
# SPDX-License-Identifier: GPL-3.0-or-later


import json

from gi.repository import GObject, Gio, Gtk, Adw, Pango

from wike.data import settings, languages


# Language links panel for sidebar

@Gtk.Template(resource_path='/com/github/hugolabe/Wike/gtk/langlinks.ui')
class LanglinksPanel(Adw.Bin):

  __gtype_name__ = 'LanglinksPanel'

  filter_dropdown = Gtk.Template.Child()
  search_entry = Gtk.Template.Child()
  langlinks_list = Gtk.Template.Child()

  # Initialize widgets and connect signals

  def __init__(self, window):
    super().__init__()

    self._window = window

    self.filter_model = Gio.ListStore(item_type=LanguagesFilter)
    self.filter_model.append(LanguagesFilter(_('Your languages'), 0))
    self.filter_model.append(LanguagesFilter(_('All languages'), 1))

    filter_factory = Gtk.SignalListItemFactory()
    filter_factory.connect('setup', self._filter_factory_setup)
    filter_factory.connect('bind', self._filter_factory_bind)

    self.filter_dropdown.set_model(self.filter_model)
    self.filter_dropdown.set_factory(filter_factory)

    self.langlinks_list.set_filter_func(self._filter_list, self.search_entry)

    settings.bind('filter-langlinks', self.filter_dropdown, 'selected', Gio.SettingsBindFlags.DEFAULT)
    settings.connect('changed::filter-langlinks', self._filter_langlinks_changed_cb)
    settings.connect('changed::show-flags', self._settings_show_flags_changed_cb)

    self.search_entry.connect('search-changed', self._search_entry_changed_cb)
    self.langlinks_list.connect('row-activated', self._list_activated_cb)

  # Setup filter item with a label

  def _filter_factory_setup(self, factory, list_item):
    label = Gtk.Label()
    label.set_xalign(0)
    label.set_ellipsize(Pango.EllipsizeMode.END)
    list_item.set_child(label)

  # Bind label property for filter item

  def _filter_factory_bind(self, factory, list_item):
    label = list_item.get_child()
    history_filter = list_item.get_item()
    history_filter.bind_property('name', label, 'label', GObject.BindingFlags.SYNC_CREATE)

  # Filter list for entry content

  def _filter_list(self, row, search_entry):
    text = search_entry.get_text()
    if text == '':
      return True

    if not row.get_activatable():
      return False

    if row.lang_name.lower().startswith(text.lower()) or row.lang.lower().startswith(text.lower()):
      return True
    else:
      return False

  # Populate langlinks list

  def populate(self, langlinks):
    while True:
      row = self.langlinks_list.get_row_at_index(0)
      if row:
        self.langlinks_list.remove(row)
      else:
        break

    if not langlinks:
      return

    user_langlinks = []
    other_langlinks = []
    for langlink in langlinks:
      if langlink['lang'] in languages.items:
        user_langlinks.append(langlink)
      else:
        other_langlinks.append(langlink)

    if user_langlinks:
      for langlink in sorted(user_langlinks, key=lambda x: x['lang']):
        row = LanglinksRow(langlink['url'], langlink['lang'], langlink['autonym'].capitalize(), langlink['*'])
        self.langlinks_list.append(row)

    if other_langlinks and settings.get_int('filter-langlinks') == 1:
      section_label = Gtk.Label()
      section_label.set_label(_('Other languages'))
      section_label.add_css_class('heading')
      section_label.set_xalign(0)
      section_label.set_margin_start(3)
      section_label.set_margin_end(3)
      row = Gtk.ListBoxRow()
      row.set_child(section_label)
      row.set_activatable(False)
      row.set_selectable(False)
      self.langlinks_list.append(row)
      for langlink in sorted(other_langlinks, key=lambda x: x['lang']):
        row = LanglinksRow(langlink['url'], langlink['lang'], langlink['autonym'].capitalize(), langlink['*'])
        self.langlinks_list.append(row)

  # Settings filter langlinks changed event

  def _filter_langlinks_changed_cb(self, settings, key):
    self.populate(self._window.page.wikiview.langlinks)

  # Settings show flags changed event

  def _settings_show_flags_changed_cb(self, settings, key):
    self.populate(self._window.page.wikiview.langlinks)

  # Refresh list on filter entry changed

  def _search_entry_changed_cb(self, search_entry):
    self.langlinks_list.invalidate_filter()

  # On list activated load page in choosen language

  def _list_activated_cb(self, langs_list, row):
    if self._window.panel_split.get_collapsed():
      self._window.panel_split.set_show_sidebar(False)

    self._window.page.wikiview.load_wiki(row.uri)


# This object represent a language filter in dropdown

class LanguagesFilter(GObject.Object):
  __gtype_name__ = 'LanguagesFilter'

  # Set values

  def __init__(self, name, level):
    super().__init__()

    self._name = name
    self._level = level

  # Name of filter property

  @GObject.Property(type=str)
  def name(self):
    return self._name

  # Id property

  @GObject.Property(type=int)
  def level(self):
    return self._level


# Row on langlinks list

@Gtk.Template(resource_path='/com/github/hugolabe/Wike/gtk/langlinks-row.ui')
class LanglinksRow(Gtk.ListBoxRow):

  __gtype_name__ = 'LanglinksRow'

  lang_label = Gtk.Template.Child()
  title_label = Gtk.Template.Child()
  flag_image = Gtk.Template.Child()

  # Set values and widgets

  def __init__(self, uri, lang, lang_name, title):
    super().__init__()

    self.uri = uri
    self.lang = lang
    self.lang_name = lang_name

    self.lang_label.set_label(lang_name)
    self.title_label.set_label(title)

    if settings.get_boolean('show-flags'):
      gfile = Gio.File.new_for_uri('resource:///com/github/hugolabe/Wike/icons/scalable/emblems/' + lang + '.svg')
      if gfile.query_exists():
        self.flag_image.set_from_icon_name(lang)
    else:
      self.flag_image.set_visible(False)



# This file is part of Wike (com.github.hugolabe.Wike)
# SPDX-FileCopyrightText: 2021-25 Hugo Olabera
# SPDX-License-Identifier: GPL-3.0-or-later


import sys

import gi
gi.require_version('Gdk', '4.0')
gi.require_version('Gtk', '4.0')
gi.require_version('Adw', '1')
gi.require_version('Soup', '3.0')
gi.require_version('WebKit', '6.0')
from gi.repository import GLib, Gio, Gdk, Gtk, Adw, WebKit

from wike.data import settings, languages, history, bookmarks
from wike.prefs import PrefsDialog
from wike.window import Window
from wike.view import network_session


# Main application class for Wike

class Application(Adw.Application):

  # Initialize app

  def __init__(self):
    super().__init__(application_id='com.github.hugolabe.Wike', flags=Gio.ApplicationFlags.HANDLES_COMMAND_LINE)

    self._window = None
    self._launch_uri = ''

    self.add_main_option('url', b'u', GLib.OptionFlags.NONE, GLib.OptionArg.STRING, 'Open Wikipedia URL', None)

  # Load custom css and set actions

  def do_startup(self):
    Adw.Application.do_startup(self)

    self._css_sepia = Gtk.CssProvider()
    self._css_sepia.load_from_resource('/com/github/hugolabe/Wike/style-sepia.css')

    self._style_manager = Adw.StyleManager.get_default()
    theme = settings.get_int('theme')
    match theme:
      case 0:
        self._style_manager.set_color_scheme(Adw.ColorScheme.FORCE_LIGHT)
      case 1:
        self._style_manager.set_color_scheme(Adw.ColorScheme.FORCE_DARK)
      case 2:
        self._style_manager.set_color_scheme(Adw.ColorScheme.FORCE_LIGHT)
        Gtk.StyleContext.add_provider_for_display(Gdk.Display.get_default(), self._css_sepia, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION)
      case 3:
        self._style_manager.set_color_scheme(Adw.ColorScheme.PREFER_LIGHT)

    action = Gio.SimpleAction.new('prefs', None)
    action.connect('activate', self._prefs_cb)
    self.add_action(action)
    self.set_accels_for_action('app.prefs', ('<Ctrl>comma',))

    action = Gio.SimpleAction.new('theme-system', None)
    action.connect('activate', self._theme_system)
    self.add_action(action)

    action = Gio.SimpleAction.new('theme-light', None)
    action.connect('activate', self._theme_light)
    self.add_action(action)

    action = Gio.SimpleAction.new('theme-sepia', None)
    action.connect('activate', self._theme_sepia)
    self.add_action(action)

    action = Gio.SimpleAction.new('theme-dark', None)
    action.connect('activate', self._theme_dark)
    self.add_action(action)

    action = Gio.SimpleAction.new('about', None)
    action.connect('activate', self._about_cb)
    self.add_action(action)

    action = Gio.SimpleAction.new('quit', None)
    action.connect('activate', self._quit_cb)
    self.add_action(action)
    self.set_accels_for_action('app.quit', ('<Ctrl>Q',))

  # Manage command line options

  def do_command_line(self, command_line):
    options = command_line.get_options_dict().end().unpack()

    if 'url' in options:
      self._launch_uri = options['url']

      if self._window:
        self._window.new_page(self._launch_uri, None, True)

    self.activate()
    return 0

  # Create main window and connect close event

  def do_activate(self):
    if not self._window:
      self._window = Window(self, self._launch_uri)
      self._window.connect('close-request',self._window_close_cb)

    self._window.present()

  # Set theme system

  def _theme_system(self, action, parameter):
    self._style_manager.set_color_scheme(Adw.ColorScheme.PREFER_LIGHT)
    Gtk.StyleContext.remove_provider_for_display(Gdk.Display.get_default(), self._css_sepia)
    settings.set_int('theme', 3)

  # Set theme light

  def _theme_light(self, action, parameter):
    self._style_manager.set_color_scheme(Adw.ColorScheme.FORCE_LIGHT)
    Gtk.StyleContext.remove_provider_for_display(Gdk.Display.get_default(), self._css_sepia)
    settings.set_int('theme', 0)

  # Set theme sepia

  def _theme_sepia(self, action, parameter):
    self._style_manager.set_color_scheme(Adw.ColorScheme.FORCE_LIGHT)
    Gtk.StyleContext.add_provider_for_display(Gdk.Display.get_default(), self._css_sepia, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION)
    settings.set_int('theme', 2)

  # Set theme dark

  def _theme_dark(self, action, parameter):
    self._style_manager.set_color_scheme(Adw.ColorScheme.FORCE_DARK)
    Gtk.StyleContext.remove_provider_for_display(Gdk.Display.get_default(), self._css_sepia)
    settings.set_int('theme', 1)

  # Show preferences dialog

  def _prefs_cb(self, action, parameter):
    prefs_dialog = PrefsDialog(self._window)
    prefs_dialog.present(self._window)

  # Show about dialog

  def _about_cb(self, action, parameter):
    builder = Gtk.Builder()
    builder.add_from_resource('/com/github/hugolabe/Wike/gtk/about.ui')
    about_dialog = builder.get_object('about_dialog')

    about_dialog.add_link(_('Source Code'), 'https://github.com/hugolabe/Wike')
    about_dialog.add_link(_('Report an Issue'), 'https://github.com/hugolabe/Wike/issues')
    about_dialog.add_link(_('Help Translating'), 'https://poeditor.com/join/project?hash=kNgJu4MAum')
    about_dialog.set_developers(['Hugo Olabera <hugolabe@proton.me>', _('Contributors') + ' https://github.com/hugolabe/Wike/graphs/contributors',])
    about_dialog.add_credit_section(_('Flag Icons'), ['HatScripts https://github.com/HatScripts/',])
    about_dialog.add_legal_section('Wikipedia', _('All content from Wikipedia.org and available under CC BY-SA 4.0 unless otherwise noted.\n\nWike is an independent development not endorsed by or affiliated with the Wikimedia Foundation.'), Gtk.License.UNKNOWN, None)
    about_dialog.add_legal_section('Circle Flags', 'Copyright © 2022 HatScripts', Gtk.License.MIT_X11, None)

    about_dialog.present(self._window)

  # On window close quit app

  def _window_close_cb(self, window):
    window.tabview.disconnect(window.handler_selpage)
    quit_action = self.lookup_action('quit')
    quit_action.activate()

  # Save settings and data and quit app

  def _quit_cb(self, action, parameter):
    if self._window.is_maximized():
      settings.set_boolean('window-max', True)
    else:
      settings.set_boolean('window-max', False)
      size = self._window.get_default_size()
      settings.set_int('window-width', size[0])
      settings.set_int('window-height', size[1])

    if self._window.page.wikiview.is_local():
      settings.set_string('last-uri', '')
    else:
      settings.set_string('last-uri', self._window.page.wikiview.get_base_uri())

    tabs_data = {}
    for i_page in range(self._window.tabview.get_n_pages()):
      tabpage = self._window.tabview.get_nth_page(i_page)
      page = tabpage.get_child()
      if page.lazy_load:
        uri = page.lazy_load[0]
        title = page.lazy_load[1]
      else:
        uri = page.wikiview.get_base_uri()
        title = page.wikiview.title
      if not uri.startswith('about'):
        tabs_data[uri] = title
    settings.set_value('last-tabs', GLib.Variant('a{ss}', tabs_data))

    if not settings.get_boolean('keep-history'):
      history.clear()

    if settings.get_boolean('clear-data'):
      data_manager = network_session.get_website_data_manager()
      data_manager.clear(WebKit.WebsiteDataTypes.ALL, 0, None, None, None)

    settings.sync()
    languages.save()
    history.save()
    bookmarks.save()

    self.quit()


# Run app

def main(version):
  app = Application()
  return app.run(sys.argv)



# This file is part of Wike (com.github.hugolabe.Wike)
# SPDX-FileCopyrightText: 2021-25 Hugo Olabera
# SPDX-License-Identifier: GPL-3.0-or-later


import os

from gi.repository import GLib, Gio, Gdk, Gtk, Adw, WebKit

from wike.data import settings
from wike.bookmarks import BookmarksPanel
from wike.history import HistoryPanel
from wike.langlinks import LanglinksPanel
from wike.menu import ArticleMenuPopover, MainMenuPopover, ViewMenuPopover
from wike.page import PageBox
from wike.search import SearchPanel
from wike.toc import TocPanel


# Main window

@Gtk.Template(resource_path='/com/github/hugolabe/Wike/gtk/window.ui')
class Window(Adw.ApplicationWindow):

  __gtype_name__ = 'Window'

  breakpoint_window_medium = Gtk.Template.Child()
  breakpoint_window_small = Gtk.Template.Child()
  breakpoint_content = Gtk.Template.Child()
  toast_overlay = Gtk.Template.Child()
  main_button = Gtk.Template.Child()
  search_button = Gtk.Template.Child()
  toc_button = Gtk.Template.Child()
  langlinks_button = Gtk.Template.Child()
  bookmarks_button = Gtk.Template.Child()
  history_button = Gtk.Template.Child()
  panel_button = Gtk.Template.Child()
  panel_split = Gtk.Template.Child()
  panel_stack = Gtk.Template.Child()
  main_button_mob = Gtk.Template.Child()
  close_panel_button_mob = Gtk.Template.Child()
  headerbar = Gtk.Template.Child()
  newtab_button = Gtk.Template.Child()
  tabs_button = Gtk.Template.Child()
  article_button = Gtk.Template.Child()
  view_button = Gtk.Template.Child()
  panel_button_mob = Gtk.Template.Child()
  article_button_mob = Gtk.Template.Child()
  view_button_mob = Gtk.Template.Child()
  tabbar = Gtk.Template.Child()
  tabview = Gtk.Template.Child()
  taboverview = Gtk.Template.Child()

  # Initialize window, set actions and connect signals

  def __init__(self, app, launch_uri):
    super().__init__(application=app)

    width = settings.get_int('window-width')
    height = settings.get_int('window-height')

    self.set_default_size(width, height)
    if settings.get_boolean('window-max'):
      self.maximize()

    self._print_settings = Gtk.PrintSettings()

    self.main_menu_popover = MainMenuPopover()
    self.main_button.set_popover(self.main_menu_popover)

    self.article_menu_popover = ArticleMenuPopover()
    self.article_button.set_popover(self.article_menu_popover)

    self.view_menu_popover = ViewMenuPopover(self)
    self.view_button.set_popover(self.view_menu_popover)

    self.search_panel = SearchPanel(self)
    search_stack_page = self.panel_stack.add_named(self.search_panel, 'search')

    self.toc_panel = TocPanel(self)
    toc_stack_page = self.panel_stack.add_named(self.toc_panel, 'toc')

    self.langlinks_panel = LanglinksPanel(self)
    langlinks_stack_page = self.panel_stack.add_named(self.langlinks_panel, 'langlinks')

    self.bookmarks_panel = BookmarksPanel(self)
    bookmarks_stack_page = self.panel_stack.add_named(self.bookmarks_panel, 'bookmarks')
    
    self.history_panel = HistoryPanel(self)
    history_stack_page = self.panel_stack.add_named(self.history_panel, 'history')

    self.page = PageBox(self, None)

    if launch_uri != '':
      tabpage = self.tabview.append(self.page)
      if launch_uri == 'notfound':
        self.page.wikiview.load_message(launch_uri)
      else:
        self.page.wikiview.load_wiki(launch_uri)
    else:
      on_start = settings.get_int('on-start-load')
      last_uri = settings.get_string('last-uri')
      if on_start != 3:
        tabpage = self.tabview.append(self.page)
        match on_start:
          case 0:
            self.page.wikiview.load_main()
          case 1:
            self.page.wikiview.load_random()
          case 2:
            if last_uri != '':
              self.page.wikiview.load_wiki(last_uri)
            else:
              self.page.wikiview.load_message('blank')
      else:
        tabs_data = settings.get_value('last-tabs').unpack()
        if len(tabs_data) > 0:
          for uri, title in tabs_data.items():
            if last_uri == uri:
              tabpage = self.tabview.append(self.page)
              self.tabview.set_selected_page(tabpage)
              self.page.wikiview.load_wiki(last_uri)
            else:
              self.new_lazy_page(uri, title, None)
          if not last_uri in tabs_data:
            tabpage = self.tabview.get_nth_page(0)
            self.tabview.set_selected_page(tabpage)
            self.page = tabpage.get_child()
            self.page.load_now()
        else:
          tabpage = self.tabview.append(self.page)
          self.page.wikiview.load_message('blank')

    settings.bind('panel-pinned', self.panel_button, 'active', Gio.SettingsBindFlags.DEFAULT)
    settings.bind('panel-page', self.panel_stack, 'visible-child-name', Gio.SettingsBindFlags.DEFAULT)

    self._set_actions(app)

    self.breakpoint_window_medium.connect('apply', self._breakpoint_medium_apply_cb)
    self.breakpoint_window_medium.connect('unapply', self._breakpoint_medium_unapply_cb)
    self.breakpoint_window_small.connect('apply', self._breakpoint_small_apply_cb)
    self.breakpoint_window_small.connect('unapply', self._breakpoint_small_unapply_cb)
    self.breakpoint_content.connect('apply', self._breakpoint_content_apply_cb)
    self.breakpoint_content.connect('unapply', self._breakpoint_content_unapply_cb)
    settings.connect('changed::hide-tabs', self._settings_hide_tabs_changed_cb)

    self.handler_selpage = self.tabview.connect('notify::selected-page', self._tabview_selected_page_cb)
    self.tabview.connect('close-page', self._tabview_close_page_cb)
    self.taboverview.connect('create-tab', self._taboverview_create_tab_cb)

    self.toc_button.connect('toggled', self._bar_selector_button_toggled_cb, 'toc')
    self.search_button.connect('toggled', self._bar_selector_button_toggled_cb, 'search')
    self.langlinks_button.connect('toggled', self._bar_selector_button_toggled_cb, 'langlinks')
    self.bookmarks_button.connect('toggled', self._bar_selector_button_toggled_cb, 'bookmarks')
    self.history_button.connect('toggled', self._bar_selector_button_toggled_cb, 'history')
    self.panel_split.connect('notify::show-sidebar', self._panel_split_show_cb)
    self.close_panel_button_mob.connect('clicked', self._close_panel_button_mob_cb)

    gesture = Gtk.GestureClick(button=0, propagation_phase=Gtk.PropagationPhase.CAPTURE)
    self.add_controller(gesture)
    gesture.connect('pressed', self._gesture_click_cb)

    self._panel_split_show_cb(self.panel_split, None)
    if settings.get_boolean('hide-tabs'):
      self._hide_tabs(True)

  # Set actions for window

  def _set_actions(self, app):
    actions = [ ('prev-page', self._prev_page_cb, ('<Alt>Left',)),
                ('next-page', self._next_page_cb, ('<Alt>Right',)),
                ('new-tab', self._new_tab_cb, ('<Ctrl>T',)),
                ('close-tab', self._close_tab_cb, ('<Ctrl>W',)),
                ('next-tab', self._next_tab_cb, ('<Ctrl>Tab',)),
                ('prev-tab', self._prev_tab_cb, ('<Shift><Ctrl>Tab',)),
                ('toggle-overview', self._toggle_overview_cb, ('F4', '<Alt>T',)),
                ('add-bookmark', self._add_bookmark_cb, ('<Ctrl>D',)),
                ('show-search', self._show_search_cb, ('F2', '<Ctrl>K',)),
                ('show-toc', self._show_toc_cb, ('<Ctrl>I',)),
                ('show-langlinks', self._show_langlinks_cb, ('<Ctrl>L',)),
                ('show-bookmarks', self._show_bookmarks_cb, ('<Ctrl>B',)),
                ('show-history', self._show_history_cb, ('<Ctrl>H',)),
                ('main-page', self._main_page_cb, ('<Alt>Home',)),
                ('random-article', self._random_article_cb, ('<Alt>R',)),
                ('open-link', self._open_link_cb, ('<Alt>L',)),
                ('reload-page', self._reload_page_cb, ('F5', '<Ctrl>R',)),
                ('search-text', self._search_text_cb, ('<Ctrl>F',)),
                ('print-page', self._print_page_cb, ('<Ctrl>P',)),
                ('open-browser', self._open_browser_cb, None),
                ('copy-link', self._copy_url_cb, ('<Ctrl>U',)) ]

    for action, callback, accel in actions:
      simple_action = Gio.SimpleAction.new(action, None)
      simple_action.connect('activate', callback)
      self.add_action(simple_action)
      if accel:
        app.set_accels_for_action('win.' + action, accel)

    pin_panel_action = Gio.SimpleAction.new_stateful('pin-panel', None, GLib.Variant.new_boolean(False))
    pin_panel_action.connect('change-state', self._pin_panel_cb)
    self.add_action(pin_panel_action)
    app.set_accels_for_action('win.pin-panel', ('F9',))

    if settings.get_boolean('panel-pinned'):
      pin_panel_action.change_state(GLib.Variant.new_boolean(True))

  # On breakpoint window medium apply

  def _breakpoint_medium_apply_cb(self, break_point):
    self.panel_split.set_collapsed(True)

    pin_panel_action = self.lookup_action('pin-panel')
    pin_panel_action.set_enabled(False)

  # On breakpoint window medium unapply

  def _breakpoint_medium_unapply_cb(self, break_point):
    pin_panel_action = self.lookup_action('pin-panel')
    pin_panel_action.set_enabled(True)

    if self.panel_button.get_active():
      self.panel_split.set_collapsed(False)

  # On breakpoint window small apply

  def _breakpoint_small_apply_cb(self, break_point):
    self.main_button.set_popover(None)
    self.main_button_mob.set_popover(self.main_menu_popover)

    pin_panel_action = self.lookup_action('pin-panel')
    pin_panel_action.set_enabled(False)

  # On breakpoint window small unapply

  def _breakpoint_small_unapply_cb(self, break_point):
    self.main_button_mob.set_popover(None)
    self.main_button.set_popover(self.main_menu_popover)

    pin_panel_action = self.lookup_action('pin-panel')
    pin_panel_action.set_enabled(True)

  # On breakpoint content apply

  def _breakpoint_content_apply_cb(self, break_point):
    self._hide_tabs(True)

    self.article_button.set_popover(None)
    self.article_button_mob.set_popover(self.article_menu_popover)
    self.view_button.set_popover(None)
    self.view_button_mob.set_popover(self.view_menu_popover)

  # On breakpoint content unapply

  def _breakpoint_content_unapply_cb(self, break_point):
    if not settings.get_boolean('hide-tabs'):
      self._hide_tabs(False)

    self.article_button_mob.set_popover(None)
    self.article_button.set_popover(self.article_menu_popover)
    self.view_button_mob.set_popover(None)
    self.view_button.set_popover(self.view_menu_popover)

  # Settings hide tabs changed event

  def _settings_hide_tabs_changed_cb(self, settings, key):
    if settings.get_boolean('hide-tabs'):
      self._hide_tabs(True)
    else:
      if self.main_button.get_visible():
        self._hide_tabs(False)

  # Hide or show tabbar and related buttons

  def _hide_tabs(self, hide):
    if hide:
      self.tabbar.set_visible(False)
      self.newtab_button.set_visible(False)
      self.tabs_button.set_visible(True)
    else:
      self.tabbar.set_visible(True)
      self.newtab_button.set_visible(True)
      self.tabs_button.set_visible(False)

  # On selector button toggled change panel view

  def _bar_selector_button_toggled_cb(self, button, name):
    if button.get_active():
      self.panel_stack.set_visible_child_name(name)
      if not self.panel_split.get_show_sidebar():
        self.panel_split.set_show_sidebar(True)
      if name=='search':
        self.search_panel.search_entry.grab_focus()

  # Panel show state changed

  def _panel_split_show_cb(self, panel_split, param):
    if panel_split.get_show_sidebar():
      panel_page = self.panel_stack.get_visible_child_name()
      match panel_page:
        case 'search':
          self.search_button.set_active(True)
        case 'toc':
          self.toc_button.set_active(True)
        case 'langlinks':
          self.langlinks_button.set_active(True)
        case 'bookmarks':
          self.bookmarks_button.set_active(True)
        case 'history':
          self.history_button.set_active(True)
    else:
      self.search_button.set_active(False)
      self.toc_button.set_active(False)
      self.langlinks_button.set_active(False)
      self.bookmarks_button.set_active(False)
      self.history_button.set_active(False)
      self.page.set_focus()

  # On close panel button clicked hide panel

  def _close_panel_button_mob_cb(self, button):
    self.panel_button_mob.set_active(False)

  # Select search in panel

  def _show_search_cb(self, action, parameter):
    self.search_button.set_active(True)

  # Select toc in panel

  def _show_toc_cb(self, action, parameter):
    self.toc_button.set_active(True)

  # Select langlinks in panel

  def _show_langlinks_cb(self, action, parameter):
    self.langlinks_button.set_active(True)

  # Select bookmarks in panel

  def _show_bookmarks_cb(self, action, parameter):
    self.bookmarks_button.set_active(True)

  # Select history in panel

  def _show_history_cb(self, action, parameter):
    self.history_button.set_active(True)

  # On pin panel action set collapsed state

  def _pin_panel_cb(self, action, parameter):
    action.set_state(parameter)
    if parameter:
      self.panel_split.set_collapsed(False)
    else:
      self.panel_split.set_collapsed(True)

  # Handle mouse clicks with buttons 8 and 9

  def _gesture_click_cb(self, gesture, n_press, x, y):
    button = gesture.get_current_button()
    prev_page_action = self.lookup_action('prev-page')
    next_page_action = self.lookup_action('next-page')

    match button:
      case 8:
        gesture.set_state(Gtk.EventSequenceState.CLAIMED)
        prev_page_action.activate()
      case 9:
        gesture.set_state(Gtk.EventSequenceState.CLAIMED)
        next_page_action.activate()

  # Create new tab with a page

  def new_page(self, uri, parent, select):
    page = PageBox(self, None)
    tabpage = self.tabview.add_page(page, parent)
    tabpage.set_live_thumbnail(True)

    match uri:
      case 'main':
        page.wikiview.load_main()
      case 'random':
        page.wikiview.load_random()
      case 'blank' | 'notfound':
        page.wikiview.load_message(uri)
      case _:
        page.wikiview.load_wiki(uri)

    if select:
      self.tabview.set_selected_page(tabpage)

    return tabpage

  # New empty tab with a title for lazy-loading

  def new_lazy_page(self, uri, title, parent):
    page = PageBox(self, [uri,title])
    tabpage = self.tabview.add_page(page, parent)
    tabpage.set_live_thumbnail(True)
    tabpage.set_title(title)

    return tabpage

  # New empty tab

  def _new_tab_cb(self, action, parameter):
    self.new_page('blank', None, True)

  # Close current tab

  def _close_tab_cb(self, action, parameter):
    tabpage = self.tabview.get_selected_page()
    self.tabview.close_page(tabpage)

  # Go to next tab

  def _next_tab_cb(self, action, parameter):
    num_pages = self.tabview.get_n_pages()
    if num_pages == 1:
      return

    tabpage = self.tabview.get_selected_page()
    if self.tabview.get_page_position(tabpage) < num_pages - 1:
      self.tabview.select_next_page()
    else:
      tabpage = self.tabview.get_nth_page(0)
      self.tabview.set_selected_page(tabpage)

  # Go to previous tab

  def _prev_tab_cb(self, action, parameter):
    num_pages = self.tabview.get_n_pages()
    if num_pages == 1:
      return

    tabpage = self.tabview.get_selected_page()
    if self.tabview.get_page_position(tabpage) > 0:
      self.tabview.select_previous_page()
    else:
      tabpage = self.tabview.get_nth_page(num_pages - 1)
      self.tabview.set_selected_page(tabpage)

  # Open/close taboverview

  def _toggle_overview_cb(self, action, parameter):
    self.taboverview.set_open(not self.taboverview.get_open())

  # On tab selected event refresh headerbar and sidebar, and force loading

  def _tabview_selected_page_cb(self, tabview, value):
    tabpage = tabview.get_selected_page()
    if not tabpage:
      return

    self.page = tabpage.get_child()
    if self.page.lazy_load:
      self.page.load_now()
    else:
      self.refresh_nav_actions(self.page.wikiview)
      self.refresh_menu_actions(self.page.wikiview.is_local())
      self.set_title(self.page.wikiview.title)
      self.toc_panel.populate(self.page.wikiview.title, self.page.wikiview.sections)
      self.langlinks_panel.populate(self.page.wikiview.langlinks)
      self.bookmarks_panel.refresh_buttons()

  # On tab closed event destroy wikiview and confirm

  def _tabview_close_page_cb(self, tabview, tabpage):
    page = tabpage.get_child()
    wikiview = page.wikiview

    if self.tabview.get_n_pages() > 1:
      page.view_stack.remove(wikiview)
      wikiview.run_dispose()
      tabview.close_page_finish(tabpage, True)
    else:
      if self.taboverview.get_open():
        self.taboverview.set_open(False)
      wikiview.load_message('blank')
      tabview.close_page_finish(tabpage, False)

    return True

  # Create new tab on taboverview button clicked

  def _taboverview_create_tab_cb(self, taboverview):
    tabpage = self.new_page('blank', None, True)
    return tabpage

  # Go to previous page

  def _prev_page_cb(self, action, parameter):
    if self.page.wikiview.can_go_back():
      self.page.wikiview.go_back()

  # Go to next page

  def _next_page_cb(self, action, parameter):
    if self.page.wikiview.can_go_forward():
      self.page.wikiview.go_forward()

  # Add page to bookmarks

  def _add_bookmark_cb(self, action, parameter):
    if not self.page.wikiview.is_local():
      uri = self.page.wikiview.get_base_uri()
      title = self.page.wikiview.title
      lang = self.page.wikiview.get_lang()
      if self.bookmarks_panel.add_bookmark(uri, title, lang):
        message = _('Bookmark added: ') + title
        self.send_notification(message)

  # Show Wikipedia main page

  def _main_page_cb(self, action, parameter):
    self.page.wikiview.load_main()

  # Show Wikipedia random article

  def _random_article_cb(self, action, parameter):
    self.page.wikiview.load_random()

  # Show open link dialog

  def _open_link_cb(self, action, parameter):
    builder = Gtk.Builder()
    builder.add_from_resource('/com/github/hugolabe/Wike/gtk/dialogs.ui')
    open_link_dialog = builder.get_object('open_link_dialog')
    link_entry = builder.get_object('link_entry')

    link_entry.connect('apply', self._link_entry_apply_cb, open_link_dialog)
    link_entry.connect('changed', self._link_entry_changed_cb, open_link_dialog)

    open_link_dialog.present(self)

  # On entry apply load link if valid

  def _link_entry_apply_cb(self, link_entry, open_link_dialog):
    text = link_entry.get_text()
    link = self.page.wikiview.is_wiki_uri(text)

    if link != '':
      open_link_dialog.close()
      self.page.wikiview.load_wiki(link)
    else:
      link_entry.set_show_apply_button(False)
      link_entry.set_show_apply_button(True)
      link_entry.add_css_class('error')


  # On entry changes check input

  def _link_entry_changed_cb(self, link_entry, open_link_dialog):
    text = link_entry.get_text()

    if len(text) == 0:
      link_entry.set_show_apply_button(False)
      link_entry.set_show_apply_button(True)

    if link_entry.has_css_class('error'):
      link_entry.remove_css_class('error')

  # Reload article view

  def _reload_page_cb(self, action, parameter):
    self.page.wikiview.reload()

  # Show text search bar

  def _search_text_cb(self, action, parameter):
    if not self.page.search_bar.get_search_mode():
      self.page.search_bar.set_search_mode(True)
    else:
      self.page.search_entry.grab_focus()

  # Print current page

  def _print_page_cb(self, action, parameter):
    print_operation = WebKit.PrintOperation.new(self.page.wikiview)
    self._print_set_settings(print_operation)

    result = print_operation.run_dialog(self)
    if result == WebKit.PrintOperationResponse.PRINT:
      handler_finished = print_operation.connect('finished', self._print_operation_finished)
      print_operation.connect('failed', self._print_operation_failed, handler_finished)

  # Set print operation settings

  def _print_set_settings(self, print_operation):
    output_uri = self._print_settings.get('output-uri')
    if output_uri:
      output_dir = os.path.dirname(output_uri)
    else:
      output_dir = 'file://' + os.path.expanduser('~')

    output_name = self.page.wikiview.title + '.pdf'
    output_name = output_name.replace('/', '-')

    output_uri = os.path.join(output_dir, output_name)
    self._print_settings.set('output-uri', output_uri)
    print_operation.set_print_settings(self._print_settings)

  # Show a notification when print operation is finished

  def _print_operation_finished(self, print_operation):
    self._print_settings = print_operation.get_print_settings()
    self.send_notification(_('Print operation completed'))

  # Show a notification when print operation failed

  def _print_operation_failed(self, print_operation, error, handler_finished):
    print_operation.disconnect(handler_finished)
    self.send_notification(_('Print operation failed'))

  # Open article in external browser

  def _open_browser_cb(self, action, parameter):
    uri = self.page.wikiview.get_base_uri()

    uri_launcher = Gtk.UriLauncher.new(uri)
    uri_launcher.launch(self, None, None, None)

  # Copy article URL to clipboard

  def _copy_url_cb(self, action, parameter):
    uri = self.page.wikiview.get_base_uri()
    clipboard = Gdk.Display.get_default().get_clipboard()
    clipboard.set(uri)

    self.send_notification(_('Link copied to clipboard'))

  # Refresh navigation actions state

  def refresh_nav_actions(self, wikiview):
    prev_page_action = self.lookup_action('prev-page')
    next_page_action = self.lookup_action('next-page')

    if wikiview.can_go_back():
      prev_page_action.set_enabled(True)
    else:
      prev_page_action.set_enabled(False)
    if wikiview.can_go_forward():
      next_page_action.set_enabled(True)
    else:
      next_page_action.set_enabled(False)

  # Refresh page actions state

  def refresh_menu_actions(self, is_local):
    search_text_action = self.lookup_action('search-text')
    print_page_action = self.lookup_action('print-page')
    open_browser_action = self.lookup_action('open-browser')
    copy_url_action = self.lookup_action('copy-link')

    if is_local:
      search_text_action.set_enabled(False)
      print_page_action.set_enabled(False)
      open_browser_action.set_enabled(False)
      copy_url_action.set_enabled(False)
    else:
      search_text_action.set_enabled(True)
      print_page_action.set_enabled(True)
      open_browser_action.set_enabled(True)
      copy_url_action.set_enabled(True)

  # Show notification with provided message

  def send_notification(self, message):
    toast = Adw.Toast.new(message)
    toast.set_timeout(3)
    self.toast_overlay.add_toast(toast)



# This file is part of Wike (com.github.hugolabe.Wike)
# SPDX-FileCopyrightText: 2021-25 Hugo Olabera
# SPDX-License-Identifier: GPL-3.0-or-later


from gi.repository import GLib, Gio, Gtk, Pango

from wike.data import settings


# Popover for main menu

class MainMenuPopover(Gtk.PopoverMenu):

  # Set menu model

  def __init__(self):
    super().__init__()

    builder_menu = Gtk.Builder()
    builder_menu.add_from_resource('/com/github/hugolabe/Wike/gtk/menu-main.ui')
    main_menu = builder_menu.get_object('main_menu')
    self.set_menu_model(main_menu)


# Popover for article menu

class ArticleMenuPopover(Gtk.PopoverMenu):

  # Set menu model

  def __init__(self):
    super().__init__()

    builder_menu = Gtk.Builder()
    builder_menu.add_from_resource('/com/github/hugolabe/Wike/gtk/menu-article.ui')
    article_menu = builder_menu.get_object('article_menu')
    self.set_menu_model(article_menu)


# Popover for view menu

class ViewMenuPopover(Gtk.PopoverMenu):

  # Set menu model and add theme switcher and zoom level widgets

  def __init__(self, window):
    super().__init__()

    self._window = window

    builder_menu = Gtk.Builder()
    builder_menu.add_from_resource('/com/github/hugolabe/Wike/gtk/menu-view.ui')
    self.view_menu = builder_menu.get_object('view_menu')
    self.set_menu_model(self.view_menu)
    
    theme_switcher = ThemeSwitcher(window)
    self.add_child(theme_switcher, 'theme_switcher')

    zoom_level = ZoomLevel()
    self.add_child(zoom_level, 'zoom_level')

    if not settings.get_boolean('custom-font'):
      pango_context = self.get_pango_context()
      font_family = pango_context.get_font_description().get_family()
      settings.set_string('font-family', font_family)

    self._add_section_font()

    actions = Gio.SimpleActionGroup()
    action = Gio.SimpleAction.new_stateful('show-flags', None, GLib.Variant.new_boolean(settings.get_boolean('show-flags')))
    action.connect('change-state', self._show_flags_cb)
    actions.add_action(action)
    action = Gio.SimpleAction.new_stateful('preview-popups', None, GLib.Variant.new_boolean(settings.get_boolean('preview-popups')))
    action.connect('change-state', self._preview_popups_cb)
    actions.add_action(action)
    action = Gio.SimpleAction.new('set-font', None)
    action.connect('activate', self._set_font_cb)
    actions.add_action(action)
    self.insert_action_group('view', actions)

  # Show/hide flags for language lists

  def _show_flags_cb(self, action, parameter):
    action.set_state(parameter)
    settings.set_boolean('show-flags', parameter)

  # Show/hide previews for article links

  def _preview_popups_cb(self, action, parameter):
    action.set_state(parameter)
    settings.set_boolean('preview-popups', parameter)

  # Show choose font dialog

  def _set_font_cb(self, action, parameter):
    font_desc = Pango.FontDescription()
    font_desc.set_family(settings.get_string('font-family'))
    font_desc.set_absolute_size(settings.get_int('font-size') * Pango.SCALE)

    font_dialog = Gtk.FontChooserDialog.new(_('Pick a Font'), self._window)
    font_dialog.set_font_desc(font_desc)
    font_dialog.set_level(Gtk.FontChooserLevel.SIZE)
    font_dialog.set_modal(True)
    font_dialog.show()
    font_dialog.connect('response', self._font_dialog_cb)

  # On dialog response set font and refresh menu

  def _font_dialog_cb(self, font_dialog, response):
    if response == Gtk.ResponseType.OK:
      font_desc = font_dialog.get_font_desc()
      font_family = font_desc.get_family()
      font_size = font_desc.get_size() / Pango.SCALE
      settings.set_boolean('custom-font', True)
      settings.set_string('font-family', font_family)
      settings.set_int('font-size', font_size)
      self.view_menu.remove(3)
      self._add_section_font()
      font_dialog.destroy()
    else:
      font_dialog.destroy()

  # Add menu section to choose font

  def _add_section_font(self):
    font_label = settings.get_string('font-family') + ' ' + str(settings.get_int('font-size'))
    font_section = Gio.Menu()
    font_section.append(font_label, 'view.set-font')
    self.view_menu.append_section(_('Font'), font_section)


# Theme switcher widget for view menu

@Gtk.Template(resource_path='/com/github/hugolabe/Wike/gtk/theme-switcher.ui')
class ThemeSwitcher(Gtk.Box):

  __gtype_name__ = 'ThemeSwitcher'

  system_button = Gtk.Template.Child()
  light_button = Gtk.Template.Child()
  sepia_button = Gtk.Template.Child()
  dark_button = Gtk.Template.Child()

  # Set button theme active and connect signals

  def __init__(self, window):
    super().__init__()

    self._window = window

    theme = settings.get_int('theme')
    match theme:
      case 0:
        self.light_button.set_active(True)
      case 1:
        self.dark_button.set_active(True)
      case 2:
        self.sepia_button.set_active(True)
      case 3:
        self.system_button.set_active(True)

    self.system_button.connect('toggled', self._system_button_toggled_cb)
    self.light_button.connect('toggled', self._light_button_toggled_cb)
    self.sepia_button.connect('toggled', self._sepia_button_toggled_cb)
    self.dark_button.connect('toggled', self._dark_button_toggled_cb)

  # On button toggle change to system theme

  def _system_button_toggled_cb(self, button):
    if button.get_active():
      app = self._window.get_application()
      theme_action = app.lookup_action('theme-system')
      theme_action.activate()

  # On button toggle change to light theme

  def _light_button_toggled_cb(self, button):
    if button.get_active():
      app = self._window.get_application()
      theme_action = app.lookup_action('theme-light')
      theme_action.activate()

  # On button toggle change to sepia theme

  def _sepia_button_toggled_cb(self, button):
    if button.get_active():
      app = self._window.get_application()
      theme_action = app.lookup_action('theme-sepia')
      theme_action.activate()

  # On button toggle change to dark theme

  def _dark_button_toggled_cb(self, button):
    if button.get_active():
      app = self._window.get_application()
      theme_action = app.lookup_action('theme-dark')
      theme_action.activate()


# Zoom level widget for view menu

@Gtk.Template(resource_path='/com/github/hugolabe/Wike/gtk/zoom-level.ui')
class ZoomLevel(Gtk.Box):

  __gtype_name__ = 'ZoomLevel'

  zoom_value_button = Gtk.Template.Child()
  zoom_in_button = Gtk.Template.Child()
  zoom_out_button = Gtk.Template.Child()

  # Set buttons status and connect signals

  def __init__(self):
    super().__init__()

    self._set_buttons()

    self.zoom_value_button.connect('clicked', self._zoom_value_clicked)
    self.zoom_in_button.connect('clicked', self._zoom_in_clicked)
    self.zoom_out_button.connect('clicked', self._zoom_out_clicked)
    settings.connect('changed::zoom-level', self._zoom_level_changed_cb)

  # Set buttons sensitivity for current zoom level value

  def _set_buttons(self):
    zoom_value = settings.get_int('zoom-level')
    self.zoom_value_button.set_label(str(zoom_value) + '%')

    if zoom_value == 100:
      self.zoom_value_button.set_sensitive(False)
      self.zoom_in_button.set_sensitive(True)
      self.zoom_out_button.set_sensitive(True)
    else:
      self.zoom_value_button.set_sensitive(True)
      if zoom_value >= 200:
        self.zoom_in_button.set_sensitive(False)
        self.zoom_out_button.set_sensitive(True)
      elif zoom_value <= 50:
        self.zoom_in_button.set_sensitive(True)
        self.zoom_out_button.set_sensitive(False)
      else:
        self.zoom_in_button.set_sensitive(True)
        self.zoom_out_button.set_sensitive(True)

  # On button clicked set zoom level to 100%

  def _zoom_value_clicked(self, zoom_value_button):
    settings.set_int('zoom-level', 100)

  # On button clicked increase zoom level

  def _zoom_in_clicked(self, zoom_in_button):
    zoom_level = settings.get_int('zoom-level')
    if zoom_level < 200:
      zoom_level += 10
      settings.set_int('zoom-level', zoom_level)

  # On button clicked reduce zoom level

  def _zoom_out_clicked(self, zoom_out_button):
    zoom_level = settings.get_int('zoom-level')
    if zoom_level > 50:
      zoom_level -= 10
      settings.set_int('zoom-level', zoom_level)

  # On zoom level changed refresh buttons

  def _zoom_level_changed_cb(self, settings, key):
    self._set_buttons()



# This file is part of Wike (com.github.hugolabe.Wike)
# SPDX-FileCopyrightText: 2021-25 Hugo Olabera
# SPDX-License-Identifier: GPL-3.0-or-later


import json, urllib.parse

from gi.repository import Soup


# Create a Soup session and set user agent

session = Soup.Session.new()
session.set_user_agent('Wike/3.2.1 (https://github.com/hugolabe)')

# Get Wikipedia random page

def get_random(lang, callback):
  endpoint = 'https://' + lang + '.wikipedia.org/w/api.php'
  params = { 'action': 'query',
             'generator': 'random',
             'grnlimit': 1,
             'grnnamespace': 0,
             'prop': 'info',
             'inprop': 'url',
             'format': 'json' }

  _request(endpoint, params, callback, None)

# Get random result from response data

def random_result(async_result):
  response = session.send_and_read_finish(async_result)
  data = response.get_data()
  result = json.loads(data)

  pages = result['query']['pages']
  page_props = list(pages.values())[0]
  uri = page_props['fullurl']
  return uri

# Search Wikipedia with a limit of responses

def search(text, lang, limit, callback):
  endpoint = 'https://' + lang + '.wikipedia.org/w/api.php'
  params = { 'action': 'opensearch',
             'search': text,
             'limit': limit,
             'namespace': 0,
             'redirects': 'resolve',
             'format': 'json' }

  if callback:
    _request(endpoint, params, callback, None)
    return

  data = _request(endpoint, params, None, None)

  if data:
    result = json.loads(data)
    if len(result[1]) > 0:
      return result[1], result[3]

  return None

# Get search results from response data

def search_result(async_result):
  response = session.send_and_read_finish(async_result)
  data = response.get_data()
  result = json.loads(data)

  if len(result[1]) > 0:
    return result[1], result[3]
  else:
    return None

# Get various properties for Wikipedia page

def get_properties(page, lang, callback, user_data):
  endpoint = 'https://' + lang + '.wikipedia.org/w/api.php'
  params = { 'action': 'parse',
             'prop': 'sections|langlinks',
             'redirects': 1,
             'page': page,
             'format': 'json' }

  _request(endpoint, params, callback, user_data)

# Get properties result from response data

def properties_result(async_result):
  response = session.send_and_read_finish(async_result)
  data = response.get_data()
  result = json.loads(data)

  return result['parse']

# Perform query to Wikipedia API with given parameters

def _request(endpoint, params, callback, user_data):
  global session

  params_encoded = urllib.parse.urlencode(params, safe='%=&|')
  message = Soup.Message.new_from_encoded_form('GET', endpoint, params_encoded)

  if callback:
    session.send_and_read_async(message, 0, None, callback, user_data)
    return
  else:
    response = session.send_and_read(message, None)

  if message.get_status() == Soup.Status.OK:
    data = response.get_data()
    return data
  else:
    return None



# This file is part of Wike (com.github.hugolabe.Wike)
# SPDX-FileCopyrightText: 2021-25 Hugo Olabera
# SPDX-License-Identifier: GPL-3.0-or-later


import urllib.parse

from gi.repository import GLib, GObject, Gio, Gdk, Gtk, Adw, WebKit

from wike import wikipedia
from wike.data import settings, languages, bookmarks


# Initialize network session with cookie manager

network_session = WebKit.NetworkSession.get_default()
cookies_file_path = GLib.get_user_data_dir() + '/cookies'
cookie_manager = WebKit.NetworkSession.get_cookie_manager(network_session)
cookie_manager.set_accept_policy(WebKit.CookieAcceptPolicy.ALWAYS)
cookie_manager.set_persistent_storage(cookies_file_path, WebKit.CookiePersistentStorage.TEXT)


# Settings and user content with custom css for wikiviews

class ViewSettings(GObject.GObject):

  __gsignals__ = { 'set-skin': (GObject.SIGNAL_RUN_FIRST, None, (bool,)) }

  web_settings = WebKit.Settings()
  user_content = WebKit.UserContentManager()
  _style_manager = Adw.StyleManager.get_default()

  # Load custom css and connect view settings signals

  def __init__(self):
    super().__init__()

    self.web_settings.set_user_agent('Mozilla/5.0 (Linux; Android; Mobile) AppleWebKit/537.36 (KHTML, like Gecko)')
    self.web_settings.set_default_font_size(settings.get_int('font-size'))
    self.web_settings.set_enable_back_forward_navigation_gestures(True)

    gfile = Gio.File.new_for_uri('resource:///com/github/hugolabe/Wike/styles/view.min.css')
    try:
      gfile_contents = gfile.load_contents(None)
    except:
      print('Can’t load view css file from resources')
      self._css_view = ''
    else:
      self._css_view = gfile_contents[1].decode('utf-8')

    gfile = Gio.File.new_for_uri('resource:///com/github/hugolabe/Wike/styles/sepia.min.css')
    try:
      gfile_contents = gfile.load_contents(None)
    except:
      print('Can’t load sepia css file from resources')
      self._css_sepia = ''
    else:
      self._css_sepia = gfile_contents[1].decode('utf-8')

    self.set_style()

    settings.connect('changed::font-family', self._settings_font_family_changed_cb)
    settings.connect('changed::font-size', self._settings_font_size_changed_cb)
    settings.connect('changed::preview-popups', self._settings_preview_popups_changed_cb)
    settings.connect('changed::theme', self._settings_theme_changed_cb)
    self._style_manager.connect('notify::dark', self._system_theme_changed_cb)

  # Settings font family changed event

  def _settings_font_family_changed_cb(self, settings, key):
    self.set_style()

  # Settings font-size changed event

  def _settings_font_size_changed_cb(self, settings, key):
    self.web_settings.set_default_font_size(settings.get_int('font-size'))

  # Settings preview popups changed event

  def _settings_preview_popups_changed_cb(self, settings, key):
    self.set_style()

  # Settings theme changed event

  def _settings_theme_changed_cb(self, settings, key):
    self.set_style()

  # System theme changed event

  def _system_theme_changed_cb(self, style_manager, dark):
    theme = settings.get_int('theme')
    if theme == 3:
      self.set_style()

  # Inject stylesheets for customize article view

  def set_style(self):
    data_manager = network_session.get_website_data_manager()
    data_manager.clear(WebKit.WebsiteDataTypes.MEMORY_CACHE , 0, None, None, None)
    self.user_content.remove_all_style_sheets()

    style_view = WebKit.UserStyleSheet(self._css_view, WebKit.UserContentInjectedFrames.ALL_FRAMES, WebKit.UserStyleLevel.USER, None, None)
    self.user_content.add_style_sheet(style_view)

    css_font = 'body,h1,h2{font-family:"' + settings.get_string('font-family') + '"!important}'
    style_font = WebKit.UserStyleSheet(css_font, WebKit.UserContentInjectedFrames.ALL_FRAMES, WebKit.UserStyleLevel.USER, None, None)
    self.user_content.add_style_sheet(style_font)

    theme = settings.get_int('theme')
    is_dark = False
    match theme:
      case 1:
        is_dark = True
      case 2:
        style_sepia = WebKit.UserStyleSheet(self._css_sepia, WebKit.UserContentInjectedFrames.ALL_FRAMES, WebKit.UserStyleLevel.USER, None, None)
        self.user_content.add_style_sheet(style_sepia)
      case 3:
        if self._style_manager.get_dark():
          is_dark = True
    self.emit('set-skin', is_dark)

    if not settings.get_boolean('preview-popups'):
      css_previews = '.mwe-popups{display:none!important}'
      style_previews = WebKit.UserStyleSheet(css_previews, WebKit.UserContentInjectedFrames.ALL_FRAMES, WebKit.UserStyleLevel.USER, None, None)
      self.user_content.add_style_sheet(style_previews)


# Create view_settings object

view_settings = ViewSettings()


# Webview for Wikipedia views

class WikiView(WebKit.WebView):

  __gsignals__ = { 'load-props': (GObject.SIGNAL_RUN_FIRST, None, ()),
                   'new-page': (GObject.SIGNAL_RUN_FIRST, None, (str,)),
                   'add-bookmark': (GObject.SIGNAL_RUN_FIRST, None, (str, str, str,)) }

  title = 'Wike'
  sections = None
  langlinks = None

  fail_uri = None

  # Initialize view with user content and web settings

  def __init__(self):
    super().__init__(settings=view_settings.web_settings, user_content_manager=view_settings.user_content)

    self.set_background_color(Gdk.RGBA(1, 1, 1, 0))
    self.set_zoom_level(settings.get_int('zoom-level') / 100)

    settings.connect('changed::zoom-level', self._zoom_level_changed_cb)
    view_settings.connect('set-skin', self._set_skin_cb)

  # Set Wikipedia light or dark skin

  def set_wikipedia_skin(self, is_dark):
    if is_dark:
      script = '''
      document.documentElement.classList.remove("skin-theme-clientpref-day");
      document.documentElement.classList.add("skin-theme-clientpref-night");
      '''
    else:
      script = '''
      document.documentElement.classList.remove("skin-theme-clientpref-night");
      document.documentElement.classList.add("skin-theme-clientpref-day");
      '''

    self.evaluate_javascript(script, -1, None, None, None)

  # Settings zoom level changed event

  def _zoom_level_changed_cb(self, settings, key):
    self.set_zoom_level(settings.get_int('zoom-level') / 100)

  # Set dark skin signal emitted

  def _set_skin_cb(self, view_settings, is_dark):
    self.set_wikipedia_skin(is_dark)

  # Load Wikipedia article by URI

  def load_wiki(self, uri):
    self.stop_loading()
    self.load_uri(uri)

  # Go to section in page

  def load_section(self, anchor):
    section = urllib.parse.quote(anchor)
    uri = self.get_uri()
    uri_elements = urllib.parse.urlparse(uri)
    if uri_elements[5] != section:
      section_uri_elements = (uri_elements[0], uri_elements[1], uri_elements[2], uri_elements[3], uri_elements[4], section)
      section_uri = urllib.parse.urlunparse(section_uri_elements)
      self.load_uri(section_uri)

  # Load Wikipedia main page

  def load_main(self):
    uri = 'https://' + settings.get_string('search-language') + '.wikipedia.org'
    self.load_wiki(uri)

  # Get Wikipedia random article async

  def load_random(self):
    wikipedia.get_random(settings.get_string('search-language'), self._on_random_finished)

  # On random finished get results

  def _on_random_finished(self, session, async_result, user_data):
    try:
      uri = wikipedia.random_result(async_result)
    except:
      self.load_message('error')
    else:
      if uri:
        self.load_wiki(uri)
      else:
        self.load_message('notfound')

  # Load void html for status page

  def load_message(self, message_type):
    uri = 'about:' + message_type
    self.load_alternate_html('', uri, None)

  # Get base uri for current article

  def get_base_uri(self):
    uri = self.get_uri()
    uri_elements = urllib.parse.urlparse(uri)
    if uri_elements[5]:
      base_uri_elements = (uri_elements[0], uri_elements[1], uri_elements[2], '', '', '')
      base_uri = urllib.parse.urlunparse(base_uri_elements)
      return base_uri
    else:
      return uri

  # Get language for current article

  def get_lang(self):
    uri = self.get_uri()
    uri_elements = urllib.parse.urlparse(uri)
    uri_scheme = uri_elements[0]
    uri_netloc = uri_elements[1]

    if uri_scheme == 'about':
      return None
    else:
      lang = uri_netloc.split('.', 1)[0]
      return lang

  # Check if current page is local (status page)

  def is_local(self):
    uri = self.get_uri()
    uri_elements = urllib.parse.urlparse(uri)
    uri_scheme = uri_elements[0]

    if uri_scheme == 'about':
      return True
    else:
      return False

  # Check if uri is a valid Wikipedia URL

  def is_wiki_uri(self, uri):
    uri_elements = urllib.parse.urlparse(uri)
    uri_scheme = uri_elements[0]
    uri_netloc = uri_elements[1]
    uri_path = uri_elements[2]

    if uri_netloc.endswith('.wikipedia.org'):
      if uri_scheme == '':
        uri_scheme = 'https'
      if uri_path == '':
        uri_path = '/'
      if uri_path.startswith('/wiki/') or uri_path == '/':
        base_uri_elements = (uri_scheme, uri_netloc, uri_path, '', '', '')
        base_uri = urllib.parse.urlunparse(base_uri_elements)
        return base_uri
      else:
        return ''
    else:
      return ''

  # Set toc, langlinks and title for current article

  def do_load_changed(self, event):
    if event != WebKit.LoadEvent.COMMITTED:
      return

    uri = self.get_uri()
    uri_elements = urllib.parse.urlparse(uri)
    uri_scheme = uri_elements[0]
    uri_netloc = uri_elements[1]
    uri_path = uri_elements[2]

    props = None
    self.title = 'Wike'
    self.sections = None
    self.langlinks = None

    if uri_scheme == 'about':
      if uri_path == 'notfound':
        self.title = _('Article not found')
      elif uri_path == 'error':
        self.title = _('Can’t access Wikipedia')
      self.emit('load-props')
      return
    else:
      page = uri_path.replace('/wiki/', '', 1)
      lang = uri_netloc.split('.', 1)[0]
      wikipedia.get_properties(page, lang, self._on_properties_finished, page)

  # On properties finished get results

  def _on_properties_finished(self, session, async_result, page):
    try:
      props = wikipedia.properties_result(async_result)
    except:
      self.emit('load-props')
    else:
      if props:
        self.title = props['title']
        self.sections = props['sections']
        self.langlinks = props['langlinks']
      else:
        title_raw = page.replace('_', ' ')
        self.title = urllib.parse.unquote(title_raw)
      self.emit('load-props')

  # Webview load error event

  def do_load_failed(self, event, uri, error):
    if error.code != 302:
      self.fail_uri = uri
      self.load_message('error')

    return True

  # Webview decision policy event

  def do_decide_policy(self, decision, decision_type):
    if decision_type == WebKit.PolicyDecisionType.NAVIGATION_ACTION or decision_type == WebKit.PolicyDecisionType.NEW_WINDOW_ACTION:
      nav_action = decision.get_navigation_action()
      nav_type = nav_action.get_navigation_type()
      mouse_button = nav_action.get_mouse_button()
      uri = nav_action.get_request().get_uri()
      uri_elements = urllib.parse.urlparse(uri)
      uri_scheme = uri_elements[0]
      uri_netloc = uri_elements[1]
      uri_path = uri_elements[2]
      uri_fragment = uri_elements[5]
      match nav_type:
        case WebKit.NavigationType.LINK_CLICKED:
          if uri_netloc.endswith('.wikipedia.org') and (uri_path.startswith('/wiki/') or uri_path == '/'):
            base_uri_elements = (uri_scheme, uri_netloc, uri_path, '', '', '')
            base_uri = urllib.parse.urlunparse(base_uri_elements)
            if mouse_button == 2:
              decision.ignore()
              self.emit('new-page', base_uri)
            else:
              if base_uri != self.get_base_uri():
                decision.ignore()
                self.load_wiki(uri)
          else:
            decision.ignore()
            Gtk.show_uri(None, uri, Gdk.CURRENT_TIME)
        case WebKit.NavigationType.RELOAD | WebKit.NavigationType.BACK_FORWARD:
          if uri_scheme == 'about':
            decision.ignore()
        case WebKit.NavigationType.OTHER:
          if uri_netloc.startswith('upload.'):
            decision.ignore()
    return True

  # Manage webview context menu

  def do_context_menu(self, menu, hit_test):
    if hit_test.context_is_editable() or hit_test.context_is_media() or hit_test.context_is_scrollbar():
      return True

    action_tab = Gio.SimpleAction.new('new-tab', GLib.VariantType('s'))
    action_tab.connect('activate', self._new_tab_cb)
    action_browser = Gio.SimpleAction.new('open-browser', GLib.VariantType('s'))
    action_browser.connect('activate', self._open_browser_cb)
    action_bookmark = Gio.SimpleAction.new('add-bookmark', GLib.VariantType('s'))
    action_bookmark.connect('activate', self._add_bookmark_cb)
    action_clipboard = Gio.SimpleAction.new('copy-link', GLib.VariantType('s'))
    action_clipboard.connect('activate', self._copy_link_cb)

    if hit_test.get_context() == WebKit.HitTestResultContext.DOCUMENT:
      base_uri = self.get_base_uri()
      uri = self.is_wiki_uri(base_uri)
      if uri == '':
        action_bookmark.set_enabled(False)

      item = WebKit.ContextMenuItem.new_separator()
      menu.append(item)
      item = WebKit.ContextMenuItem.new_from_gaction(action_bookmark, _('Add Bookmark'), GLib.Variant.new_string(uri))
      menu.append(item)

    if hit_test.context_is_link() and not hit_test.context_is_image():
      link = hit_test.get_link_uri()
      uri = self.is_wiki_uri(link)
      if uri == '':
        action_bookmark.set_enabled(False)
        action_tab.set_enabled(False)

      if uri != '':
        link = uri
      item = WebKit.ContextMenuItem.new_from_gaction(action_tab, _('Open Link in New Tab'), GLib.Variant.new_string(uri))
      menu.append(item)
      item = WebKit.ContextMenuItem.new_from_gaction(action_browser, _('Open Link in Browser'), GLib.Variant.new_string(link))
      menu.append(item)
      item = WebKit.ContextMenuItem.new_separator()
      menu.append(item)
      item = WebKit.ContextMenuItem.new_from_gaction(action_bookmark, _('Add Link to Bookmarks'), GLib.Variant.new_string(uri))
      menu.append(item)
      item = WebKit.ContextMenuItem.new_from_gaction(action_clipboard, _('Copy Link to Clipboard'), GLib.Variant.new_string(link))
      menu.append(item)

    allowed_actions = (WebKit.ContextMenuAction.NO_ACTION,
                       WebKit.ContextMenuAction.CUSTOM,
                       WebKit.ContextMenuAction.GO_BACK,
                       WebKit.ContextMenuAction.GO_FORWARD,
                       WebKit.ContextMenuAction.RELOAD,
                       WebKit.ContextMenuAction.COPY,
                       WebKit.ContextMenuAction.COPY_IMAGE_TO_CLIPBOARD,
                       WebKit.ContextMenuAction.COPY_IMAGE_URL_TO_CLIPBOARD)

    items = menu.get_items()
    for item in items:
      if not item.get_stock_action() in allowed_actions:
        menu.remove(item)

    return False

  # On context menu activated open link in new tab

  def _new_tab_cb(self, action, parameter):
    uri = parameter.get_string()
    self.emit('new-page', uri)

  # On context menu activated open link in browser

  def _open_browser_cb(self, action, parameter):
    uri = parameter.get_string()
    Gtk.show_uri(None, uri, Gdk.CURRENT_TIME)

  # On context menu activated add bookmark

  def _add_bookmark_cb(self, action, parameter):
    uri = parameter.get_string()
    uri_elements = urllib.parse.urlparse(uri)
    uri_netloc = uri_elements[1]
    uri_path = uri_elements[2]

    page = uri_path.replace('/wiki/', '', 1)
    title_raw = page.replace('_', ' ')
    title = urllib.parse.unquote(title_raw)
    lang = uri_netloc.split('.', 1)[0]

    self.emit('add-bookmark', uri, title, lang)

  # On context menu activated copy link to clipboard

  def _copy_link_cb(self, action, parameter):
    uri = parameter.get_string()
    clipboard = Gdk.Display.get_default().get_clipboard()
    clipboard.set(uri)
